// A Java version for the Kaser Interpreter - File IO methods
// For some of my own fun!
// Stephen M. Bennett
// August 2002


package stephen.bennett.kint;

import java.io.*;
import java.util.*;


// here we go
public class FileIO {
  // constants folder types for buildPath
  public static final int FOLDER_HOME     = 0;
  public static final int FOLDER_UI       = 1;
  public static final int FOLDER_IMAGES   = 2;
  public static final int FOLDER_BACKGNDS = 3;
  public static final int FOLDER_MUSIC    = 4;
  public static final int FOLDER_SOUNDS   = 5;

  // the home directory, set at KINT starts up
  private static String homeDir = "";

  // the file handles currently in use, index into Vector is file handle
  private static Vector fileHandles = new Vector();

  // the findfirst results currently in use, index+1 is handle (i.e. first is 1)
  private static Vector findFirstHandles = new Vector();



  // set the home Dir based on given filename
  public static void setHomeDir(String filename) {
    // get the parent
    String p = new File(filename).getParent();

    // if null then in current directory
    if (p == null) {
      homeDir = "";
    }
    else {
      // parent is home dir
      homeDir = p+File.separator;
    }

    Log.fine("home dir", homeDir);
  }


  // build the path up based on folder code
  // isUICall of true will add graphics type to files in game home folder
  public static String buildPath(String filename, int folder, boolean isUICall) {
    Log.fine("buildPath folder", folder);

    // System.out.println("buildPath filename is "+filename);

    // going to uppercase
    filename = filename.toUpperCase();
    Log.fine("buildPath filename in uppercase", filename);

    String result = "UNKNOWN";
    
    // string buf to build up results
    StringBuffer sb = new StringBuffer(homeDir);

    switch(folder) {
    case FOLDER_HOME:
      // game home
      sb.append(filename);
      if (isUICall) {
        // for the easter egg, this can be a call for file without .BMP on
        // we need to worry about about .PNG as well, as the EKSSIE will be
        // a BMP file
        if (UI.usingPNG()) {
          // just append the .PNG
          UI.appendGraphicsFileType(sb);
        }
        else {
          // if the file exists with out the extension, then we have it
          // otherwise, add the extension on
          File f = new File(sb.toString());
          if (!f.exists()) {
            UI.appendGraphicsFileType(sb);
          }
        }
      }
      break;

    case FOLDER_UI:
      // UI directory, including .BMP or .PNG
      sb.append("UI");
      sb.append(File.separator);
      sb.append(filename);
      UI.appendGraphicsFileType(sb);
      break;

    case FOLDER_IMAGES:
      // IMAGES directory, including .BMP or .PNG
      sb.append("IMAGES");
      sb.append(File.separator);
      sb.append(filename);
      UI.appendGraphicsFileType(sb);
      break;

    case FOLDER_BACKGNDS:
      // BACKGNDS directory, including .BMP or .PNG
      sb.append("BACKGNDS");
      sb.append(File.separator);
      sb.append(filename);
      UI.appendGraphicsFileType(sb);
      break;

    case FOLDER_MUSIC:
      // MUSIC directory (one level up), including .MID
      sb.append("..");
      sb.append(File.separator);
      sb.append("MUSIC");
      sb.append(File.separator);
      sb.append(filename);
      sb.append(".MID");
      break;

    case FOLDER_SOUNDS:
      // SOUNDS are in home folder, including .WAV
      sb.append(filename);
      sb.append(".WAV");
      break;

    default:
      Log.errorExit("FileIO.buildPath unknown folder code", folder);
    }

    result = sb.toString();
    Log.fine("buildPath result", result);

    // System.out.println("buildPath result: "+result);

    return result;
  }


  // open a file, returning file handle, mode bits are
  //            _O_RDONLY      0x0000  /* open for reading only */
  //            _O_WRONLY      0x0001  /* open for writing only */
  //            _O_RDWR        0x0002  /* open for reading and writing */
  //            _O_APPEND      0x0008  /* writes done at eof */
  //            _O_CREAT       0x0100  /* create and open file */
  //            _O_TRUNC       0x0200  /* open and truncate */
  //            _O_TEXT        0x4000  /* file mode is text (translated) */
  //            _O_BINARY      0x8000  /* file mode is binary (untranslated) */
  // Return -1 if error
  public static int open(String filename, int mode) {
    Log.fine("FileIO.open mode", mode);

    // using RandomAccessFile, seems to have all we need

    // don't know what to do if _O_APPEND set at the moment, so exit
    if ((mode & 0x0008) != 0) {
      Log.errorExit("FileIO.open cannot cope with _O_APPEND bit set for file "+filename);
    }

    // If _O_CREAT not given and file does not exist, return -1
    // RandomAccessFile always tries to create the file, so we do check exists
    if ((mode & 0x0100) == 0) {
      Log.fine("FileIO.open O_CREAT not given, checking exists");
      if (!(new File(filename)).exists()) {
        Log.fine("FileIO.open file does not exist", filename);
        return -1;
      }
    }

    // we don't distinguish between _O_TEXT and _O_BINARY
    // I hope that is okay
    
    // start with read mode
    String modeString = "r";

    // if _O_WRONLY or _O_RDWR given then add "w"
    // we don't distinguish between these two
    if ((mode & 0x0001) != 0 || (mode & 0x00002) != 0) {
      modeString = modeString+"w";
    }

    // create the RandomAccess file
    try {
      RandomAccessFile raf = new RandomAccessFile(filename, modeString);

      // truncate it if required
      if ((mode & 0x0200) != 0) {
        Log.fine("FileIO.open truncating file to length 0");
        raf.setLength(0);
      }

      // put it in file handles, returning handle
      int fileHandle = Utils.addObjectToHandles(raf, fileHandles, false,
                                                "fileHandles");
      return fileHandle;
    }
    catch (FileNotFoundException e1) {
      Log.severe("FileIO.open failed to open ", filename);
      Log.severe("FileIO.open mode was ", mode);
      Log.severe("FileIO.open exception ", e1.toString());
    }
    catch (IOException e2) {
      Log.severe("FileIO.open failed to truncate ", filename);
      Log.severe("FileIO.open mode was ", mode);
      Log.severe("FileIO.open exception ", e2.toString());
    }

    // failed if get to here
    return -1;
  }



  // Return 1 if file is at EOF, 0 if not, -1 on error
  public static int eof(int fileHandle) {
    // get the Random Access File
    RandomAccessFile raf = 
      (RandomAccessFile)Utils.getObjectByHandle(fileHandle, fileHandles, false, 
                                                "fileHandles");

    // convert IO exceptions to -1
    try {
      // we don't have EOF directly, but compute using current position and
      // file length
      long currentPos = raf.getFilePointer();
      long length = raf.length();
      Log.fine("FileIO.eof currentPos length", currentPos, length);

      if (currentPos >= length) {
        // at EOF
        return 1;
      }
      else {
        // not at EOF
        return 0;
      }
    }
    catch (IOException e) {
      Log.severe("IOException in FileIO.eof: "+e);
      return -1;
    }
  }




  // read n bytes from give file handle into given memory
  // return how many read, or -1 for error
  public static int read(int fileHandle, byte[] theMem, int offset, int n) {
    // get the Random Access File
    RandomAccessFile raf = 
      (RandomAccessFile)Utils.getObjectByHandle(fileHandle, fileHandles, false, 
                                                "fileHandles");

    // convert any read errors to -1 return
    try {
      // keep reading until no more or problem or got all
      Log.fine("FileIO.read", n);
      int totalRead = 0;
      while(n > 0) {
        int nread = raf.read(theMem, offset, n);
        if (nread == -1) {
          // eof reach, cannot read any more
          break;
        }

        // keep counts and offsets up to date
        totalRead += nread;
        offset += nread;
        n -= nread;
      }

      // return how many we read
      Log.fine("FileIO.read total read", totalRead);
      return totalRead;
    }
    catch (IOException e) {
      Log.severe("IOException in FileIO.read: "+e);
      return -1;
    }
  }



  // write n bytes to give file handle from given memory
  // return how many written, or -1 for error
  public static int write(int fileHandle, byte[] theMem, int offset, int n) {
    // get the Random Access File
    RandomAccessFile raf = 
      (RandomAccessFile)Utils.getObjectByHandle(fileHandle, fileHandles, false, 
                                                "fileHandles");

    // convert any errors to -1 return
    try {
      // A RandomAccessFile writes all given, we do not need to loop incase a
      // smaller ammount is written
      Log.fine("FileIO.write", n);
      raf.write(theMem, offset, n);
      return n;
    }
    catch (IOException e) {
      Log.severe("IOException in FileIO.write: "+e);
      return -1;
    }
  }



  // seek to given offset, origin is 0 for absolute, 1 for from current,
  // and 2 for from end ????
  // result is new offset, or -1 for error
  // we lose precision from long to int (but windows version does as well)
  public static int seek(int fileHandle, int offset, int origin) {
    // get the Random Access File
    RandomAccessFile raf = 
      (RandomAccessFile)Utils.getObjectByHandle(fileHandle, fileHandles, false,
                                                "fileHandles");

    Log.fine("FileIO.seek offset origin", offset, origin);

    // IOException taken as error
    try {
      // Java only has absolute set, so need to convert origin
      long newOffset = offset;

      if (origin == 1) {
        // add in the current position
        long currentPos = raf.getFilePointer();
        Log.fine("FileIO.seek adding currentPos to offset", currentPos);
        newOffset += currentPos;
      }
      else if(origin == 2) {
        // add in the length
        long len = raf.length();
        Log.fine("FileIO.seek adding length to offset", len);
        newOffset += len;
      }

      // do the seek
      raf.seek(newOffset);

      // find and return current position
      long newPos = raf.getFilePointer();
      Log.fine("FileIO.seek new position", newPos);
      return (int)newPos;
    }
    catch (IOException e) {
      Log.fine("IOException in FileIO.seek: "+e);
    }

    return -1;
  }



  // close the file whose file handle is given
  // no return
  public static void close(int fileHandle) {
    // get the Random Access File
    RandomAccessFile raf = 
      (RandomAccessFile)Utils.getObjectByHandle(fileHandle, fileHandles, false,
                                                "fileHandles");

    // close it, ignoring errors
    try {
      raf.close();
    }
    catch (IOException e) {
      Log.fine("ignoring IOException in FileIO.close");
    }

    // no longer want this file handle (I assume)
    Log.fine("FileIO.close freeing file handle", fileHandle);
    Utils.removeObjectByHandle(fileHandle, fileHandles, false, "fileHandles");
  }



  // Return the size of a file, -1 if does not exist or error getting info
  public static int size(String filename) {
    Log.fine("FileIO.size", filename);

    // create a file for it
    File f = new File(filename);

    // does it exist
    if (f.exists()) {
      // return its size [don't worry about loss of precision, KINT.EXE does not]
      long s = f.length();
      Log.fine("FileIO.size", s);
      return (int)s;
    }

    Log.fine("FileIO.size file does not exist");
    return -1;
  }


  // Delete the named file, no return required
  public static void delete(String filename) {
    Log.fine("FileIO.delete", filename);

    // create a file for it
    File f = new File(filename);

    // just do it
    f.delete();
  }


  /////////////////////////////////////////
  // Find First and Next code
  /////////////////////////////////////////
  
  // find the first file according to the give filename spec
  // need to remember list of results to support findNext
  public static FindFirstFiles findFirst(String fileSpec) {
    // System.out.println("FileIO.findFirst file spec "+fileSpec);
    Log.fine("FileIO.findFirst file spec", fileSpec);

    // for now I am assuming this is [<path....><last sep>]*.<ext>
    // I will improve if/when a request that is not of this format is given
    
    // find the last . for the extension
    int lastDot = fileSpec.lastIndexOf('.');
    if (lastDot == -1) {
      Log.errorExit("FileIO.findFirst cannot cope with file spec of "+fileSpec);
    }

    // remember the ext (including the .)
    String ext = fileSpec.substring(lastDot);
    Log.fine("FileIO.findFirst ext", ext);

    // construct the filter for this extension
    WithExtFilenameFilter filter = new WithExtFilenameFilter(ext);

    // create a file for the (optional) path 
    String path = null;
    int lastSep = fileSpec.lastIndexOf(File.separator);
    if (lastSep == -1) {
      // only *.<ext> given, use local dir
      path = ".";
    }
    else {
      path = fileSpec.substring(0, lastSep);
    }
    Log.fine("FileIO.findFirst path", path);
    // System.out.println("FileIO.findFirst path \""+path+"\"");

    // check rest is exactly "*"
    String rest = null;
    if (lastSep == -1) {
      rest = fileSpec.substring(0, lastDot);
    }
    else {
      rest = fileSpec.substring(lastSep+1, lastDot);
    }
    if (!rest.equals("*")) {
      Log.errorExit("FileIO.findFirst cannot cope with file spec of "+fileSpec);
    }

    // get the list of files matching filename spec
    File[] files = (new File(path)).listFiles(filter);

    // if this is empty then we return null
    if (files == null || files.length == 0) {
      Log.fine("FileIO.findFirst no matching files");
      return null;
    }

    // construct the object that holds this result
    FindFirstFiles foundFiles = new FindFirstFiles(files);

    // remember it, so it can be used in findNext, and tell the object its handle
    int handle = Utils.addObjectToHandles(foundFiles, findFirstHandles, true,
                                          "findFirstHandles");
    foundFiles.setHandle(handle);

    return foundFiles;
  }


  // move the given find first result on to next file, return null if no more
  public static FindFirstFiles findNext(int handle) {
    Log.fine("FileIO.findNext handle", handle);

    FindFirstFiles foundFiles = 
               (FindFirstFiles)Utils.getObjectByHandle(handle, findFirstHandles,
                                                       true, "findFirstHandles");

    // try to move to next file
    if (!foundFiles.increment()) {
      // not possible
      Log.fine("FileIO.findNext no more files left");
      return null;
    }

    // okay
    return foundFiles;
  }



  // close the named find first result
  public static void findClose(int handle) {
    Log.fine("FileIO.findClose handle", handle);

    // just remove it, and let gc do its work
    Utils.removeObjectByHandle(handle, findFirstHandles, true, 
                               "findFirstHandles");
  }



  // useful function to drop filename extension, i.e. everything from last .
  // onwards
  public static String dropFilenameExt(String name) {
    Log.fine("dropping filename extension", name);

    // find last .
    int i = name.lastIndexOf('.');

    if (i == -1) {
      // extension not present
      return name;
    }

    // return everything up to, but not including .
    return name.substring(0, i);
  }


  // Our class to keep files only of the given extension (case insensitive like windows)
  static class WithExtFilenameFilter implements FilenameFilter {
    // our extension
    private String ext;

    // constructor
    WithExtFilenameFilter(String e) {
      // remember the extension, lowercase
      ext = e.toLowerCase();
    }

    // the interface method, return true if accept this file
    public boolean accept(File dir, String name) {
      // System.out.println("accept: "+name);
      return name.toLowerCase().endsWith(ext);
    }
  }




  // Return our CVS id
  public static String getCVSID() {
    return "$Id: FileIO.java,v 1.4 2002/11/28 19:32:31 Bennett Stephen Exp $";
  }


}


/*
 * $Log: FileIO.java,v $
 * Revision 1.4  2002/11/28 19:32:31  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.3  2002/10/12 14:25:22  Bennett Stephen
 * working version
 *
 * Revision 1.2  2002/10/06 15:30:36  Bennett Stephen
 * safe version, still being worked on
 *
 * Revision 1.1  2002/09/08 19:38:23  Bennett Stephen
 * intial version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
