// A Java version for the Kaser Interpreter
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started

package stephen.bennett.kint;

// our imports
import java.io.*;
import javax.swing.*;


public class KInt {
  // revision constants
  private final static int INTERP_MAJOR_REV = 2;
  private final static int INTERP_MINOR_REV = 3;

  // KE EVENT OFFSETS
  public final static int KEEVENT_SCREEN_CHANGE = 60;   // 0x3c 
  public final static int KEEVENT_ALPHAKEY_DOWN = 64;   // 0x40    
  public final static int KEEVENT_ALPHAKEY_UP   = 68;   // 0x44 
  public final static int KEEVENT_BEGIN_STARTUP = 72;   // 0x48 
  public final static int KEEVENT_END_STARTUP   = 76;   // 0x4c 
  public final static int KEEVENT_WINDOW_HIDDEN = 80;   // 0x50         
  public final static int KEEVENT_WINDOW_MOVE   = 84;   // 0x54         
  public final static int KEEVENT_WINDOW_SIZE   = 88;   // 0x58         
  public final static int KEEVENT_POINTER_MOVE  = 92;   // 0x5c 
  public final static int KEEVENT_KEY_DOWN      = 96;   // 0x60         
  public final static int KEEVENT_KEY_UP        = 100;  // 0x64 
  public final static int KEEVENT_CHAR          = 104;  // 0x68 
  public final static int KEEVENT_TIMER         = 108;  // 0x6c 
  public final static int KEEVENT_IDLE          = 112;  // 0x70 
  public final static int KEEVENT_MUSIC_DONE    = 116;  // 0x74         
  public final static int KEEVENT_QUERY_QUIT    = 120;  // 0x78         
  public final static int KEEVENT_SHUT_DOWN     = 124;  // 0x7c



  // the main registers
  // 32*4 + 48*2 + 32*1 = 256 (32 DWORDS, 48 WORDS, 32 BYTES)
  private static byte[] mainRegs = new byte[256]; 

  // the main Kode - initialised as .KE file is loaded
  private static byte[] mainKode;

  // the stack base - initialised as .KE file is loaded
  private static int stackBase;

  // a gate on the keEvent to prevent re-entry
  private static boolean gateEvent = false;

  // indicate if kill program has been initiated
  private static int killpgm = 0;

  // indicate when program ready, so timer events can be sent
  private static boolean programReady = false;

  // for my performance measurements
  private static long perfStartTime;



  // here we go
  public static void main(String[] args) {
    System.out.println("Java Kaser Interpreter, v1.0, Copyright 2002 Stephen M. Bennett");
    System.out.println("Based on the Kaser Interpreter, Copyright 2002 Everett Kaser");

    perfStartTime = System.currentTimeMillis();

    // process the params
    String keFileName = null;
    for(int i=0; i<args.length; i++) {
      String arg = args[i];

      if (arg.equals("-version")) {
        // -version print out version numbers
        System.out.println(KInt.getCVSID());
        System.out.println(CallKEIdleRunnable.getCVSID());
        System.out.println(ExecKE.getCVSID());
        System.out.println(FileIO.getCVSID());
        System.out.println(FindFirstFiles.getCVSID());
        System.out.println(KComponentListener.getCVSID());
        System.out.println(KKeyListener.getCVSID());
        System.out.println(KMouseListener.getCVSID());
        System.out.println(KMouseMotionListener.getCVSID());
        System.out.println(KPanel.getCVSID());
        System.out.println(KTests.getCVSID());
        System.out.println(KTimerListener.getCVSID());
        System.out.println(KWindowListener.getCVSID());
        System.out.println(Log.getCVSID());
        System.out.println(Mem.getCVSID());
        System.out.println(Music.getCVSID());
        System.out.println(Printing.getCVSID());
        System.out.println(SysFuncs.getCVSID());
        System.out.println(TextState.getCVSID());
        System.out.println(UI.getCVSID());
        System.out.println(Utils.getCVSID());
        System.out.println(WindowsBMP.getCVSID());
        System.exit(0);
      }
      else if (arg.equals("-info")) {
        // set logging to info level
        Log.setLevelInfo();
      }
      else if (arg.equals("-fine")) {
        // set logging to fine level
        Log.setLevelFine();
      }
      else if (arg.equals("-debugOnKey")) {
        // debug info when "I" is pressed, and fine when "F" is pressed
        KKeyListener.setDebugOnKey();
      }
      else if (arg.equals("-directdraw")) {
        // use direct draw graphic ops
        // NOT really needed now as I have the double buffering correct
        UI.useDirectDraw();
      }
      else if (arg.equals("-png")) {
        // use PNG files
        UI.usePNGFiles();
      }
      else if (arg.equals("-tests")) {
        // for testing, next arg is the test name (has to be there :-)
        KTests.runTests(args[i+1]);
        System.exit(0);
      }
      else {
        // must be the filename, check we have not got it yet
        if (keFileName != null) {
          System.err.println("Usage: KInt [-version] [<file.ke>]");
          System.exit(-1);
        }
        else {
          keFileName = arg;
        }
      }
    }

    // open the logging file
    Log.openLog("jkint.log");

    // default to MAIN.KE if not supplied
    if (keFileName == null) {
      keFileName = "MAIN.KE";
    }

    // remember home directory of this file (for loading other files relative)
    FileIO.setHomeDir(keFileName);

    // load the KE file
    loadKEFile(keFileName);

    // set start point for progressive time count
    SysFuncs.setStartTime();

    // init the screen
    UI.initScreen();

    // create the main window - and show for us
    // we do it early as BEING_STARTUP may want to write to it
    UI.createWindow();

    // call begin startup code in the KE module
    Log.info("Calling KEEVENT_BEGIN_STARTUP");
    keEvent(KEEVENT_BEGIN_STARTUP, 0, 0, 0);

    // the Java Swing rule is:
    // Once a Swing component has been realized, all code that might affect or 
    // depend on the state of that component should be executed in the 
    // event-dispatching thread. 

    // Therefore any calls to keEvent should now be done in the
    // event-dispatching thead.  I will do via invokeAndWait().

    // call the end start up code in the KE module
    Runnable callEndStartUp = new Runnable() {
        public void run() { 
          Log.info("Calling KEEVENT_END_STARTUP");
          keEvent(KEEVENT_END_STARTUP, 0, 0, 0);
        }
      };
    try {
      SwingUtilities.invokeAndWait(callEndStartUp);
    }
    catch (Exception e) {
      Log.errorExit("Caught exception waiting for callEndStartUp", e);
    }


    // show the window  
    // ** not for Java as we had to realise the window at creation so that
    // ** the image could be created
    // showWindow();

    // send one IDLE message to the KASM code in case it needs to do anything 
    // once right up front after everything else has been initialized.
    callKEIdle(0);

    // program now ready, timer events can fire
    programReady = true;


    // this is the main "message pump" loop.  While no messages are waiting for 
    // the program and the IDLE call to the KASM code returns non-0, it keeps 
    // sending IDLE messages to the KASM code.  If the KASM code returns 0, 
    // then no more IDLE messages are sent until some Windows message 
    // (Java listened-for event) message arrives for the program.

    // For java we don't have central Get/Translate/Dispatch point
    // We can peek the event queue but those items are dealt with elsewhere
    // An exact duplicate of the windows algorithm is not possible (without
    // taking control of the event dispatcher), so I go for what I see as
    // the essence of the Windows's .KE IDLE idea.  I hope the differences
    // are not important!?

    // The algorithm
    // While the event queue is empty, keeping calling KE IDLE code until
    // it returns non-zero.  Then wait for an event (for the case where
    // queue was empty and KE IDLE code returned 0) and then wait for queue 
    // empty before calling KE IDLE.

    // There is a potential problem here, being a time gap between the different
    // checks.  It is possible that an event can appear and be processed (hence
    // removed from the queue) so this code could actually have to wait
    // for another event before KE IDLE code is called again.  I assume this
    // will not be too disasterous!?

    // Also I can find no way to wait for an empty event queue, except to sleep
    // and keep checking.  Currently I sleep for 1 second at a time, so calling
    // of IDLE may have to wait up to a second, compared to the Windows 
    // algorithm.  I could reduce the sleep time, but that would make the program
    // more busy during idle time.  I could also take over the event dispatcher
    // but you should only do that if you are very sure of what you are doing.

    while (true) {
      int idleCount = 0;
      boolean queueIsEmpty = false;

      // while event queue empty, call idle until it returns non-zero
      while ((queueIsEmpty = UI.isAWTEventQueueEmpty()) && 
             callKEIdle(idleCount++) != 0) {
        // do nothing
      }

      // wait for an event - for the case where queue was empty and IDLE returns
      // 0 (that is KE is not interested in doing anything at the current IDLE
      // moment) so we wait for something (i.e. some event) to happen
      if (queueIsEmpty) {
        UI.waitForAWTEventToHappen();
      }

      // Now wait for the empty AWT queue (i.e. until we are IDLE again)
      UI.waitForEmptyAWTEventQueue();
    }
  }


  // call the idle event and return what it return
  // for Java this has to be called in the event dispatch thread
  private static int callKEIdle(int idleCount) {
    int result = -1;

    CallKEIdleRunnable doit = new CallKEIdleRunnable(idleCount);
    try {
      SwingUtilities.invokeAndWait(doit);
      result = doit.getResult();
    }
    catch (Exception e) {
      Log.errorExit("Caught exception waiting for KEEVENT_IDLE to return", e);
    }

    Log.fine("callKEIdle returning ", result);
    return result;
  }



  // indicate that program is ready (just for timer event handling at moment)
  public static boolean isProgramReady() {
    return programReady;
  }



  // load the KE file
  private static void loadKEFile(String filename) {
    Log.info("Loading KE File", filename);

    try {
      // open the file
      FileInputStream fis = new FileInputStream(filename);

      // read the header
      byte[] fileheader = new byte[20];
      int nRead = fis.read(fileheader);
      if (nRead != 20) {
        Log.errorExit("Could only read "+nRead+" of file header, needed 20");
      }

      // check its a .KE file - look for KASMB
      if (!new String(fileheader, 0, 5).equals("KASMB")) {
        Log.errorExit("Not a .KE file");
      }

      // check version number of the file
      int majorVersion = fileheader[6];
      int minorVersion = fileheader[7];
      Log.fine("File requires KaserInterpreter version", majorVersion, minorVersion);
      if (majorVersion > INTERP_MAJOR_REV ||
          (majorVersion == INTERP_MAJOR_REV && 
           minorVersion > INTERP_MINOR_REV)) {
        Log.errorExit("This .KE file requires a new KaserInterpreter");
        // System.out.println("WARNING!!!! This .KE file requires a new KaserInterpreter, but will continue anyway");
      }

      // get the initialized code size
      int initCodeSize = Mem.GETMEMD(fileheader, 8);
      Log.fine("initCodeSize", initCodeSize);

      // get uninitialized code size
      int uninitCodeSize = Mem.GETMEMD(fileheader, 12);
      Log.fine("uninitCodeSize", uninitCodeSize);

      // now we know the stack base
      stackBase = initCodeSize+uninitCodeSize;
      Log.fine("stackBase", stackBase);

      // get size of stack
      int stackSize = Mem.GETMEMD(fileheader, 16);
      Log.fine("stackSize", stackSize);

      // now we know the memory size
      int memorySize = initCodeSize+uninitCodeSize+stackSize;
      Log.fine("memorySize", memorySize);

      // get memory for .KE - we use allocateMem so the handle is value
      int mainKodeMemoryHandle = Mem.allocateMem(memorySize);

      // we know it is 1!!! but check anyway
      if (mainKodeMemoryHandle != 1) {
        Log.errorExit("Allocated kode memory not handle 1, rather", 
                      mainKodeMemoryHandle);
      }

      mainKode = Mem.getMemByHandle(mainKodeMemoryHandle);

      // copy in first 20 bytes already read
      System.arraycopy(fileheader, 0, mainKode, 0, 20);

      // read the rest in 
      int remains = initCodeSize-20;
      int ptr = 20;
      while(remains != 0) {
        nRead = fis.read(mainKode, ptr, remains);
        if (nRead == -1 || nRead == 0) {
          Log.errorExit("Could not read "+remains+" bytes from file");
        }
        Log.fine("Number of bytes read from .KE file", nRead);
        ptr += nRead;
        remains -= nRead;
      }
      fis.close();

      // the zero data will be zero-ed (java defined for byte initialisation)
      // the main regs will be zero-ed (java defined for byte initialisation)
    }
    catch (Exception e) {
      Log.errorExit("Caught exception loading KE file", e);
    }

    Log.info("Loaded KE File", filename);
  }



  // invoke a .KE event given offset in Kode and three args (R15, R14, R13)
  // public as called from event listeners
  // exit program if exception caught
  public static int keEvent(int cmd, int arg1, int arg2, int arg3) {
    try {
      // seeing how much time is spent in each timer event code
      // long startTime = System.currentTimeMillis();

      // if (cmd == KEEVENT_QUERY_QUIT) {
      //   System.out.println("keEvent QUERY_QUIT entered");
      // }

      Log.info("keEvent in, cmd", cmd);
      Log.fine("keEvent thread is ", Thread.currentThread().getName());

      // check the gate (only works if single-threaded - which we are!)
      if (gateEvent) {
        return 0;
      }
      gateEvent = true;

      // R0 to code base (which is always memory handle 1 for us - was mainKode)
      Mem.SETREGD(mainRegs, Mem.REG_0, 1);  

      // init PC
      Mem.SETREGD(mainRegs, Mem.REG_3, Mem.GETMEMD(mainKode, cmd));
      Log.DWORD("REG_3", mainRegs, Mem.REG_3);

      // init stack when starting up
      if (cmd == KEEVENT_BEGIN_STARTUP) {
        Mem.SETREGD(mainRegs, Mem.REG_4, stackBase);
        Log.DWORD("REG_4", mainRegs, Mem.REG_4);

        Mem.SETREGD(mainRegs, Mem.REG_5, stackBase);
        Log.DWORD("REG_5", mainRegs, Mem.REG_5);
      }

      // put the cmd in REG_32
      Mem.SETREGW(mainRegs, Mem.REG_32, (short)cmd);
      Log.WORD("REG_32", mainRegs, Mem.REG_32);

      // put the args in REG 15, 14, 13
      Mem.SETREGD(mainRegs, Mem.REG_15, arg1);
      Log.DWORD("REG_15", mainRegs, Mem.REG_15);
      Mem.SETREGD(mainRegs, Mem.REG_14, arg2);
      Log.DWORD("REG_14", mainRegs, Mem.REG_14);
      Mem.SETREGD(mainRegs, Mem.REG_13, arg3);
      Log.DWORD("REG_13", mainRegs, Mem.REG_13);

      // call the K interpreter
      int retval = ExecKE.execKE(mainRegs, mainKode);
      Log.info("execKE returns", retval);

      // good place to do process a paint for outstanding invalid paint areas
      UI.paintInvalidArea();

      // switch off the gate
      gateEvent = false;

      // if Kode did a TERM, kill the program, but ONLY do it ONCE!
      // (The Kode may call TERM more than once due to "shut down" Kode calls)
      if (killpgm == 1) {
        // change the variable so this is not done again
        killpgm |= 2;

        // close the UI window, this raises the WindowClosed event, so the
        // same code as the User Clicking on "close window" is executed
        UI.destroyWindow();
      }

      Log.info("keEvent out, cmd", cmd);

      // if (cmd == KEEVENT_QUERY_QUIT) {
      //   System.out.println("keEvent QUERY_QUIT exited");
      // }


      // long endTime = System.currentTimeMillis();
      // System.out.println("Event "+cmd+" time is "+(endTime-startTime)+" ms");

      return retval;
    }
    catch (Exception e) {
      Log.errorExit("Caught exception processing KE event "+cmd, e);
    }
    return 0;
  }


  // remember we need to kill the program
  public static void rememberKillProgram() {
    killpgm |= 1;       
  }


  // kill the program
  public static void killProgram() {
    // printPerfInfo();
    // ExecKE.printUnusedOps();

    // System.out.println("KillProgram exiting everything now");
    System.exit(0);
  }


  // print main registers - for testing
  public static void printMainRegisters() {
    Log.registers(mainRegs);
  }


  // for performace
  public static void printPerfInfo() {
    System.out.println("PERF: Total time="+(System.currentTimeMillis()-perfStartTime));
    System.out.println("PERF: loadBitmap "+UI.perfLoadBitmap()+" ms");
    System.out.println("PERF: loadMBitmap "+UI.perfLoadMBitmap()+" ms");
  }



  // Return our CVS id
  public static String getCVSID() {
    return "$Id: KInt.java,v 1.7 2003/01/21 15:48:04 Bennett Stephen Exp $";

  }

}

/*
 * $Log: KInt.java,v $
 * Revision 1.7  2003/01/21 15:48:04  Bennett Stephen
 * support for debug on key flag and v1.0 welcome message
 *
 * Revision 1.6  2002/11/28 19:32:30  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.5  2002/10/12 14:24:35  Bennett Stephen
 * simple changes
 *
 * Revision 1.4  2002/10/06 15:30:23  Bennett Stephen
 * moved to CVS methods
 *
 * Revision 1.3  2002/09/08 19:43:20  Bennett Stephen
 * set home directory and some more versioning
 *
 * Revision 1.2  2002/08/10 08:23:07  Bennett Stephen
 * version, info and fine params added
 *
 * Revision 1.1  2002/08/07 18:47:18  Bennett Stephen
 * initial version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
