// A Java version for the Kaser Interpreter - play Music and Sounds
// For some of my own fun!
// Stephen M. Bennett
// 29Jun02 Started


package stephen.bennett.kint;

import java.io.*;
import java.util.*;
import javax.sound.sampled.*;
import javax.sound.midi.*;
import javax.swing.*;


public class Music implements MetaEventListener {
  // we will remember sound clips, as there will only be few of them
  private static Map clips = new HashMap();

  // remember if music is currently play, and event says when it is stopped
  // or explicit method call to stop it
  private static boolean musicPlaying = false;

  // we keep a static ref to the Midi sequencer
  private static Sequencer sequencer;
  

  // play WAV sound based on given filename
  public static void playSound(String filename) {
    // find the full path to the sound file 
    String wavFile = FileIO.buildPath(filename, FileIO.FOLDER_SOUNDS, false);
    Log.fine("Playing sound ", wavFile);

    try {
      // see if we already have it
      Clip clip = (Clip)clips.get(wavFile);

      if (clip != null) {
        Log.fine("Using existing clip for this sound");

        // just rewind, so we can play again
        clip.setFramePosition(0);
      }
      else {
        // will have to load from file

        // get the audio input stream
        AudioInputStream ais = AudioSystem.getAudioInputStream(new File(wavFile));

        // get the audio format
        AudioFormat format = ais.getFormat();

        // define the type of line we need
        DataLine.Info info = new DataLine.Info(Clip.class, format);

        // get the clip
        clip = (Clip) AudioSystem.getLine(info);

        // reserve it for us
        clip.open(ais);

        // remember it
        clips.put(wavFile, clip);
      }

      // okay play it
      clip.start();
    }
    catch (Exception e) {
      Log.severe("Caught exception in playSound: "+e);
    }
  }


  // play MIDI music based on given filename
  public static void playMusic(String filename) {
    // find the full path to the music file 
    String midiFile = FileIO.buildPath(filename, FileIO.FOLDER_MUSIC, false);

    // get the sequencer if not done already
    if (sequencer == null) {
      // we need to catch the Midi Unavailable exception
      try {
        sequencer = MidiSystem.getSequencer();

        // an instance of us will be told when music stops
        sequencer.addMetaEventListener(new Music());
      }
      catch (MidiUnavailableException e) {
        Log.severe("Exception opening midi system: "+e);

        // no point in progressing
        return;
      }
    }

    try { 
      // open the sound file as a sequence
      Sequence sound = MidiSystem.getSequence(new File(midiFile));

      // open the sequencer now (okay to re-open?)
      sequencer.open();

      // set the sequence to play
      sequencer.setSequence(sound);

      // start playing
      sequencer.start();

      // remember we are playing
      musicPlaying = true;

      Log.info("music now playing");
    }
    catch (MidiUnavailableException e1) {
      Log.severe("Exception opening sequencer: "+e1);
    }
    catch (InvalidMidiDataException e2) {
      Log.severe("Exception loading midi file "+midiFile+" : "+e2);
    }
    catch (IOException e3) {
      Log.severe("Exception loading midi file "+midiFile+" : "+e3);
    }
  }


  // stop the MIDI music
  public static void stopMusic() {
    // Log.info("stopMusic: musicPlaying="+musicPlaying);
    // Log.info("stopMusic: sequencer="+sequencer);

    if (musicPlaying) {
      // set this to false first, as seq.stop will cause event asynchronously
      musicPlaying = false;

      // try to stop the sequence - don't worry about exceptions
      if (sequencer != null) {
        try {
          sequencer.stop();
          sequencer.close();
        }
        catch (Exception e) {
          // ignore this
          Log.fine("Exception stopping/closing sequence: "+e);
        }
      }
      Log.info("music stopped");
    }
  }


  // is the MIDI music playing?
  // return 0 if no music, !=0 if music is playing
  public static short isMusicPlaying() {
    Log.fine("isMusicPlaying", musicPlaying);

    if (musicPlaying) {
      return 1;
    }
    else {
      return 0;
    }
  }


  // For the Meta Event Listener for the sequencer
  // need to be careful here, as meta comes in on another thread
  // and we don't want to raise the event if we have just called stopMusic
  // ourselves.  So (first attempt) is that stopMusic sets musicPlaying
  // to false before stopping the music, and we only call KEEVENT_MUSIC_DONE
  // if we think the music is playing when the end of track message arrives
  public void meta(MetaMessage message) {
    // Log.info("meta: message="+message);

    if (message.getType() == 47) {  // 47 is end of track
      // we are only interested in end of track, if we think its playing
      if (musicPlaying) {
        stopMusic();

        // raise .KE function for music stopped
        // this has to be done in the UI thread, as music sequencer has its own
        // thread (see comments in KInt.java)
        Runnable callMusicDone = new Runnable() {
            public void run() { 
              Log.info("Calling KEEVENT_MUSIC_DONE");
              KInt.keEvent(KInt.KEEVENT_MUSIC_DONE, 0, 0, 0);
            }
          };
        try {
          SwingUtilities.invokeAndWait(callMusicDone);
        }
        catch (Exception e) {
          Log.errorExit("Caught exception waiting for callMusicDone", e);
        }
      }
    }
  }



  // Return our CVS id
  public static String getCVSID() {
    return "$Id: Music.java,v 1.4 2003/01/21 15:45:27 Bennett Stephen Exp $";
  }


}



/*
 * $Log: Music.java,v $
 * Revision 1.4  2003/01/21 15:45:27  Bennett Stephen
 * don't do KEEVENT_MUSIC_DONE when program stops the music
 *
 * Revision 1.3  2002/10/12 14:22:48  Bennett Stephen
 * working version
 *
 * Revision 1.2  2002/10/06 15:25:24  Bennett Stephen
 * safe version, still being worked on
 *
 * Revision 1.1  2002/08/07 19:22:20  Bennett Stephen
 * Initial Version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
