// A Java version for the Kaser Interpreter - printing routines
// For some of my own fun!
// Stephen M. Bennett
// 09Oct02 Started

package stephen.bennett.kint;

import java.awt.*;


public class Printing {
  // keep the print job that is running
  private static PrintJob printJob;

  // the graphics for the current page being printed
  private static Graphics printGraphics;



  // start the printing - true if print started, false otherwise
  public static boolean startPrinting() {
    Log.fine("Printing.startPrinting");

    // print job's come from the UI frame
    printJob = UI.getPrintJob();

    // check we got one - will be null if user cancelled?
    if (printJob == null) {
      Log.fine("print job is null");
      return false;
    }

    return true;
  }


  // stop the printing - should only be called if print job defined
  public static void stopPrinting() {
    Log.fine("Printing.stopPrinting");

    // dispose of any graphics if currently defined
    if (printGraphics != null) {
      printGraphics.dispose();
      printGraphics = null;
    }

    // stop the print job if defined
    if (printJob != null) {
      printJob.end();
      printJob = null;
    }
  }


  // get the page size - should only be called if print job defined
  public static Dimension getPageSize() {
    if (printJob == null) {
      Log.severe("Printing.getPageSize called with no print job");
      return null;
    }

    return printJob.getPageDimension();
  }
  

  // get the resolution - should only be called if print job defined
  public static int getResolution() {
    if (printJob == null) {
      Log.severe("Printing.getPrinterResolution called with no print job");
      return 0;
    }

    return printJob.getPageResolution();
  }


    

  // begin next page - should only be called if print job defined
  // return success
  public static boolean beginPage() {
    Log.fine("Printing.beginPage");

    if (printJob == null) {
      Log.severe("Printing.beginPrinting called with no print job");
      return false;
    }

    if (printGraphics != null) {
      Log.severe("Printing.beginPrinting called before previous endPage");
      return false;
    }


    // start the page, by getting the graphics
    printGraphics = printJob.getGraphics();

    // check we got one
    if (printGraphics == null) {
      Log.fine("Printing.beginPage print job graphics is null");
      return false;
    }

    return true;
  }


  // end next page - should only be called if have a current page
  // return success
  public static boolean endPage() {
    Log.fine("Printing.endPage");

    if (printGraphics == null) {
      Log.severe("Printing.endPrinting called with no print job graphics");
      return false;
    }

    // page ends by disposing of graphics
    printGraphics.dispose();
    printGraphics = null;

    return true;
  }


  // print the given text - should only be called if print graphics defined
  public static void printText(String s, int c, int x, int y, short fontHandle) {
    Log.fine("Printing.printText", s);
    Log.fine("colour", c);
    Log.fine("x y", x, y);
    Log.fine("font handle", fontHandle);

    if (printGraphics == null) {
      Log.severe("Printing.printText called with no print job graphics");
      return;
    }

    // set the colour
    printGraphics.setColor(UI.getColorFromBGR(c));

    // set the font
    Font font = UI.getFont(fontHandle);
    printGraphics.setFont(font);

    // convert y to Java y
    int newy = UI.kintTextY2Java(y, font, printGraphics);

    // draw the string
    printGraphics.drawString(s, x, newy);
  }



  // print the given lines - should only be called if print graphics defined
  public static void drawPolyline(int c, int[] xs, int[] ys, int nPoints) {
    Log.fine("Printing.drawPolyline", nPoints);
    Log.fine("colour", c);

    if (printGraphics == null) {
      Log.severe("Printing.drawPolyline called with no print job graphics");
      return;
    }

    // set the colour
    printGraphics.setColor(UI.getColorFromBGR(c));

    // draw a poly line
    printGraphics.drawPolyline(xs, ys, nPoints);
  }



  // print the named bitmap - should only be called if print graphics defined
  public static void printBitmap(int handle, 
                                 int xDst, int yDst, int wDst, int hDst,
                                 int xSrc, int ySrc, int wSrc, int hSrc) {
    Log.fine("Printing.printBitmap", handle);
    Log.fine("xDst yDst", xDst, yDst);
    Log.fine("wSrc hSrc", wSrc, hSrc);
    Log.fine("xSrc ySrc", xSrc, ySrc);
    Log.fine("wSrc hSrc", wSrc, hSrc);

    if (printGraphics == null) {
      Log.severe("Printing.printBitmap called with no print job graphics");
      return;
    }

    // get the image to print by its handle
    Image image = UI.getImage(handle);

    // draw the image, scaling as required
    // sizes are the same so no actual scalling takes place)
    boolean complete = printGraphics.drawImage(image, 
                                   xDst, yDst, xDst+wDst, yDst+hDst,
                                   xSrc, ySrc, xSrc+wSrc, ySrc+hSrc,
                                   null);

    // if not completely drawn then help!
    if (!complete) {
      Log.errorExit("Printing.printBitmap did not complete");
    }
  }



  // get text size for named text
  public static int getTextSize(String s, Dimension d, short fontHandle) {
    Log.fine("Printing.getTextSize", s);
    Log.fine("font handle", fontHandle);

    if (printGraphics == null) {
      // exit here as Java version differs from KINT which only assumes
      // that SYS_PRINTSTART has been called
      Log.errorExit("Printing.getTextSize called with no print job graphics");
    }

    // call the common code in UI for this font, and our print graphics
    return UI.getTextSize(s, d, UI.getFont(fontHandle), printGraphics);
  }



  // Return our CVS id
  public static String getCVSID() {
    return "$Id: Printing.java,v 1.2 2002/11/28 19:32:29 Bennett Stephen Exp $";
  }


}


/*
 * $Log: Printing.java,v $
 * Revision 1.2  2002/11/28 19:32:29  Bennett Stephen
 * release candidate 1 version
 *
 * Revision 1.1  2002/10/10 19:50:14  Bennett Stephen
 * initial version
 *
 *
 */

// ----------------------------------------------------
// Emacs magic - please do not remove the following lines
/*
 * Local Variables:
 * write-contents-hooks: ((lambda () (untabify (point-min) (point-max))))
 * End:
*/
// ----------------------------------------------------
